from __future__ import print_function
import argparse
import os
from termcolor import colored
from scandb.models import init_db
from scandb.nmap import import_nmap_file
from scandb.nessus import import_nessus_file

def importer():
    """
    Entry point for the console script scandb-importer. This script allows to import either a single nessus|nmap XML-file or
    several nessus|nmap XML-files within a given directory.
    :return:
    """
    parser = argparse.ArgumentParser(description="")
    parser.add_argument("--db", type=str, required=False, default="scandb.sqlite")
    parser.add_argument("--file", metavar="FILE", type=str, default=None, nargs="*",
                        help="The nessus and/or nmap file(s)")
    parser.add_argument("--dir", metavar="DIR", type=str, default=None,
                        help="Directory name with nessus and/or nmap files")
    args = parser.parse_args()

    db = args.db
    filename = args.file
    dir = args.dir

    # initialize the database
    database = init_db(db)

    if filename is None and dir is None:
        # either a filename or a directory must be specified
        parser.print_usage()
        return

    if filename is not None:
        # import a single nessus/nmap XML-file
        for file in filename:
            if file.endswith('.nessus'):
                import_nessus_file(file)
            if file.endswith('.xml'):
                import_nmap_file(file)
    if dir is not None:
        # import several nessus/nmap files within a directory
        for filename in os.listdir(dir):
            if filename.endswith('.nessus'):
                fullname = os.path.join(dir, filename)
                import_nessus_file(fullname)
            if filename.endswith('.xml'):
                fullname = os.path.join(dir, filename)
                import_nmap_file(fullname)

    database.close()



import hashlib

BUF_SIZE = 65536


def hash_file(filename):
    """
    Calculate a SHA-512 hash from the given file.
    :param filename: File from which the hash should be calculated
    :return:
    """
    sha512 = hashlib.sha512()

    with open(filename, 'rb') as f:
        while True:
            data = f.read(BUF_SIZE)
            if not data:
                break
            sha512.update(data)

    return sha512.hexdigest()


def get_open_ports(host, protocol='tcp'):
    ports = []
    for port, proto in host.get_open_ports():
        if protocol == proto:
            ports.append(port)
    return ports


def get_ports(host):
    """
    This function generates a list of open ports. Each entry is a tuple which contains the port, protocol, servicename,
    state and a banner string.

    :param host: The NmapHost object
    :type host libnmap.objects.host.NmapHost

    :return: list of tuples
    """
    ports = []
    for port, proto in host.get_ports():
        if host.get_service(port, proto) is not None:
            service = host.get_service(port, proto)
            servicename = service.service
            state = service.state
            banner = service.banner
            ports.append((port, proto, servicename, state, banner))
    return ports


def get_hostname(host):
    """
    This function get the hostname of the given host.

    :param host: The NmapHost object
    :type host: libnmap.objects.host.NmapHost

    :return: hostname
    :rtype: str
    """
    hostname = ""
    for name in host.hostnames:
        if name == "localhost" and hostname != "":
            continue
        hostname = name
    return hostname


def get_best_os_match(host):
    os_matches = host.os_match_probabilities()
    os = ""
    os_gen = ""
    if len(os_matches) > 0:
        os = os_matches[0].name
        if len(os_matches[0].osclasses) >0:
            os_gen = os_matches[0].osclasses[0].osgen
    return os, os_gen


def host_to_tupel(host):
    hostname = get_hostname(host)
    os, os_gen = get_best_os_match(host)
    return host.address, hostname, os, os_gen, host.status


import argparse
import sqlite3


def get_scan_stats(db):
    """
    This function creates a list with statistics for each scan that has been imported by scandb-importer.

    :param db: sqlite database created by scandb-importer
    :return: list with statistics for each scan
    """
    conn = sqlite3.connect(db)
    cur = conn.cursor()
    cur.execute("SELECT id,start,end,elapsed,hosts_total,hosts_up,hosts_down,name FROM scan")
    rows = cur.fetchall()
    conn.close()
    return rows


def handle_scan_stats(db, outfile, write_file = False, delimeter= ";"):
    results_csv = []
    results_cli = []
    fmt = '{0:^10}{1:^15}{2:^15}{3:^12}{4:^12}{5:^12}{6:^15}{7}'
    results_cli.append(fmt.format("scan id", "Start", "End", "Elapsed", "Hosts total", "Hosts up", "Hosts down",
                                  "Parameters"))
    results_csv.append(delimeter.join(["scan id", "Start", "End", "Elapsed", "Hosts total", "Hosts up", "Hosts down",
                                      "Parameters"]))

    stats = get_scan_stats(db)
    for s in stats:
        id, start, end, elapsed, total, up, down, params = (s)
        # on Nessus scans hosts up and hosts down are 'None'
        # convert numbers and None to strings to ensure that they don't case exceptions when used in format() or join()
        id = str(id); up = str(up); down = str(down); total = str(total)
        results_cli.append(fmt.format(id, start, end, elapsed, total, up, down, params))
        results_csv.append(delimeter.join([id, start, end, elapsed, total, up, down, params]))

    print("\n".join(results_cli))
    if write_file:
        filename = "{0}-scan-statistics.csv".format(outfile)
        with open(filename, "w") as f:
            f.write("\n".join(results_csv))


def get_vuln_stats(db):
    """
    This function creates a list with a number of vulnerabilities per host and severity. The static is created across
    all imported Nessus scans but each Nessus Plugin-ID is only counted once per host.

    :param db: sqlite database created by scandb-importer
    :return: list with number of vulnerabilities per host
    """
    sql = "SELECT address,\
    COUNT(CASE WHEN severity = 4 THEN 1 END) as CRITICAL, \
    COUNT(CASE WHEN severity = 3 THEN 1 END) as HIGH,\
    COUNT(CASE WHEN severity = 2 THEN 1 END) as MEDIUM,\
    COUNT(CASE WHEN severity = 1 THEN 1 END) as LOW,\
    COUNT(CASE WHEN severity = 0 THEN 1 END) as INFO\
    from \
    ( select distinct address, plugin, severity   from host h left join vuln v on h.id = v.host_id )\
    GROUP by address\
    order by CRITICAL DESC, HIGH DESC, MEDIUM DESC, LOW DESC, INFO DESC;"
    conn = sqlite3.connect(db)
    cur = conn.cursor()
    cur.execute(sql)
    rows = cur.fetchall()
    conn.close()
    return rows


def handle_vuln_stats(db, outfile, write_file = False, delimeter=";"):
    results_csv = []
    results_cli = []
    fmt = '{0:>20}{1:>15}{2:>15}{3:>15}{4:>15}{5:>15}'
    results_cli.append(fmt.format("Address", "Critical", "High", "Medium", "Low", "Info"))
    results_csv.append(delimeter.join(["Address", "Critical", "High", "Medium", "Low", "Info"]))

    stats = get_vuln_stats(db)

    for s in stats:
        address, c, h, m, l, i = (s)
        # convert counter values to strings to ensure that they don't case exceptions when used in format() or join()
        c = str(c); h = str(h); m = str(m); l = str(l); i = str(i)
        results_cli.append(fmt.format(address, c, h, m, l, i))
        results_csv.append(delimeter.join([address, c, h, m, l, i]))

    print("\n".join(results_cli))
    if write_file:
        filename = "{0}-vuln-statistics.csv".format(outfile)
        with open(filename, "w") as f:
            f.write("\n".join(results_csv))


def get_port_stats(db):
    """
    This function creates a list with a number of open TCP- and UDP-Ports per host. The static is created across
    all imported Nmap scans (only ports with status 'open' are counted).

    :param db: sqlite database created by scandb-importer
    :return: list with number of TCP- and UDP-Ports per host
    """
    sql = "select address,\
    COUNT(CASE WHEN protocol = 'tcp' THEN 1 END) as TCP,\
    COUNT(CASE WHEN protocol = 'udp' THEN 1 END) as UDP\
    from (select DISTINCT address, port, protocol from port where status = 'open')\
    GROUP by address"
    conn = sqlite3.connect(db)
    cur = conn.cursor()
    cur.execute(sql)
    rows = cur.fetchall()
    conn.close()
    return rows


def handle_port_stats(db, outfile, write_file = False, delimeter=";"):
    results_csv = []
    results_cli = []
    fmt = '{0:>20}{1:>15}{2:>15}'
    results_cli.append(fmt.format("Address", "TCP", "UDP"))
    results_csv.append(delimeter.join(["Address", "TCP", "UDP"]))

    stats = get_port_stats(db)
    for s in stats:
        address, tcp, udp = (s)
        # convert counter values to strings to ensure that they don't case exceptions when used in format() or join()
        tcp = str(tcp); udp = str(udp)
        results_cli.append(fmt.format(address, tcp, udp))
        results_csv.append(delimeter.join([address, tcp, udp]))

    print("\n".join(results_cli))
    if write_file:
        filename = "{0}-port-statistics.csv".format(outfile)
        with open(filename, "w") as f:
            f.write("\n".join(results_csv))


host_port_list = """
    select address , group_concat(distinct port), protocol from port where protocol = 'tcp' and status='open' group by address
    union
    select address , group_concat(distinct port), protocol from port where protocol = 'udp' and status='open' group by address;"""


def gen_host_port_list(db,outfile):
    conn = sqlite3.connect(db)
    cur = conn.cursor()
    cur.execute(host_port_list)
    rows = cur.fetchall()
    filename = "{0}-hostportlist.csv".format(outfile)
    with open(filename,'w') as f:
        for r in rows:
            f.write(";".join(r))
            f.write("\n")
    print ("Results written to : {0}".format(filename))
    conn.close()


def statistics_cli():
    parser = argparse.ArgumentParser(description="")
    parser.add_argument("--db", type=str, required=False, default="scandb.sqlite")
    parser.add_argument("-s", "--scan-statistics", required=False, action='store_true', default=False,
                        help="Print statistics for each scan")
    parser.add_argument("-v", "--vuln-statistics", required=False, action='store_true', default=False,
                        help="Print number of vulns foreach host.")
    parser.add_argument("-p", "--port-statistics", required=False, action='store_true', default=False,
                        help="Print number of 'open' TCP and UDP ports foreach host.")
    parser.add_argument("--host-portlist", required=False, action='store_true', default=False,
                        help="generate a csv with a list of TCP and UDP Ports per host")
    parser.add_argument("-d", "--delimeter", required=False, type=str,  default=";", help="Delimeter for CSV files.")
    parser.add_argument("-o", "--outfile", required=False, default="scandb", help="Prefix for output files.")
    parser.add_argument("-w", "--write-file", required=False, action='store_true', default=False,
                        help="Write data to CSV file. Prefix of filename can be changed with parameter outfile")

    args = parser.parse_args()

    do_stats = False
    if args.scan_statistics or args.vuln_statistics or args.port_statistics:
        do_stats = True

    if not do_stats and not args.host_portlist:
        parser.print_usage()
        return

    if args.scan_statistics:
        handle_scan_stats(args.db, args.outfile, args.write_file, args.delimeter)

    if args.vuln_statistics:
        handle_vuln_stats(args.db, args.outfile, args.write_file, args.delimeter)

    if args.port_statistics:
        handle_port_stats(args.db, args.outfile, args.write_file, args.delimeter)

    if args.host_portlist:
        gen_host_port_list(args.db, args.outfile)


import argparse
import sqlite3


def get_host_list(db, status):
    conn = sqlite3.connect(db)
    cur = conn.cursor()
    cur.execute("SELECT distinct address FROM host WHERE status like ? ;", (status,))
    rows = cur.fetchall()
    ips = [ x[0] for x in rows]
    conn.close()
    return ips


def get_host_list_by_both(db, tcp, udp):
    conn = sqlite3.connect(db)
    cur = conn.cursor()
    sql = "SELECT distinct address FROM port WHERE port in ({seq_tcp}) and protocol = 'tcp' " \
          "or port in ({seq_udp})  and protocol ='udp'".format(
            seq_tcp=','.join(['?'] * len(tcp)),
            seq_udp=','.join(['?'] * len(udp)))
    cur.execute(sql, (tcp,udp,))
    rows = cur.fetchall()
    ips = [x[0] for x in rows]
    conn.close()
    return ips


def get_host_list_by_udp(db, udp):
    conn = sqlite3.connect(db)
    cur = conn.cursor()
    sql = "SELECT distinct address FROM port where port in ({seq_udp}) and protocol = 'udp'".format(
            seq_udp=','.join(['?'] * len(udp)))
    cur.execute(sql, udp)
    rows = cur.fetchall()
    ips = [x[0] for x in rows]
    conn.close()
    return ips


def get_host_list_by_tcp(db, tcp):
    conn = sqlite3.connect(db)
    cur = conn.cursor()
    sql = "SELECT distinct address FROM port where port in ({seq_tcp}) and protocol = 'tcp'".format(
        seq_tcp=','.join(['?'] * len(tcp)))
    cur.execute(sql, tcp)
    rows = cur.fetchall()
    ips = [x[0] for x in rows]
    conn.close()
    return ips


def services_cli():
    parser = argparse.ArgumentParser(description="")
    parser.add_argument("--db", type=str, required=False, default="scandb.sqlite")
    parser.add_argument("--status", type=str, required=False, default="up",
                        help="Status string stored in database (default: up)")
    parser.add_argument("-t", "--tcp", metavar="PORTS", required=False, type=str, default=None,
                        help="TCP ports")
    parser.add_argument("-u", "--udp", metavar="PORTS", required=False, type=str, default=None,
                        help="UDP ports")
    parser.add_argument("-o", "--operation", metavar="UNION|INTERSECTION", required=False, type=str, default="UNION",
                        help="Operation to combine the sets of TCP and UDP ports (default: UNION)")
    parser.add_argument("--list", required=False, action='store_true', default=False,
                        help="Generate a target list")
    parser.add_argument("-d", "--list-delimiter", required=False, default="\n",
                        help="Delimiter used to separate hosts in the list output")
    parser.add_argument("--list-file", metavar="FILE", required=False, type=str, default=None,
                        help="Generate a file with the targets instead of printing them to stdout")
    args = parser.parse_args()

    if not args.list and args.list_file is None:
        parser.print_usage()
        return

    ips = []

    if args.tcp is None and args.udp is None:
        ips = get_host_list(args.db, args.status)
    else:
        tcp_ports = []
        udp_ports = []
        if args.tcp is not None:
            tcp_ports = get_host_list_by_tcp(args.db, args.tcp.split(","))
        if args.udp is not None:
            udp_ports = get_host_list_by_udp(args.db, args.udp.split(","))

        if args.operation.upper() == 'INTERSECTION':
            ips = set(tcp_ports).intersection(udp_ports)
        else:
            ips = set(tcp_ports).union(udp_ports)

    if args.list:
        print(args.list_delimiter.join(ips))

    if args.list_file:
        with open(args.list_file, 'w') as f:
            f.write("\n".join(ips))





import argparse
import os
import peewee
from libnmap.parser import NmapParser
from termcolor import colored
from scandb.util import host_to_tupel, get_ports, hash_file
from scandb.models import Scan, Host, Port, init_db


def import_nmap_file(infile):
    """
    This function is responsable for importing the given file.  For each file a SHA-512 hash is calculated to ensure
    that the file is only imported once.

    :param infile: nmap XML-file to import
    :return:
    """
    sha512 = ""
    print(colored("[*] Importing file: {0}".format(infile), 'green'))
    try:
        report = NmapParser.parse_fromfile(infile) # read and parse the nmap XML file
        # calculate a SHA-512 hash. This is used to ensure that the file will not be imported more than once.
        sha512 = hash_file(infile)

        # create the database entry for the scan.
        scan = Scan(file_hash=sha512, name=report.commandline, type='nmap', start=report.started, end=report.endtime,
                    elapsed=report.elapsed, hosts_total=report.hosts_total, hosts_up=report.hosts_up,
                    hosts_down=report.hosts_down)
        scan.save()
        # import all hosts and ports present in the report
        for h in report.hosts:
            address, hostname, os, osgen, status = host_to_tupel(h)
            host = Host(address=address, hostname=hostname, os=os, os_gen=osgen, status=status, scan=scan)
            host.save()
            ports = get_ports(h)
            for p, proto, servicename, state, banner in ports:
                port = Port(host=host, address=address, port=p, protocol=proto, service=servicename, banner=banner,
                            status=state)
                port.save()

        print(colored("[*] File imported. ",'green'))

    except peewee.IntegrityError as e:
        # This error is throw when the SHA-512 hash is already present in the database. Therefore the file cannot be
        # imported again.
        print(colored("[-] File already imported: {0}".format(infile), 'red'))
    except Exception as e:
        # Invalid file format
        print(colored("[-] File cannot be imported : {0}".format(infile), 'red'))




# -*- coding: utf-8 -*-
from setuptools import setup
from codecs import open
from os import path

here = path.abspath(path.dirname(__file__))

long_description = ""
# Get the long description from the README file
with open(path.join(here, 'README.md'), encoding='utf-8') as f:
    long_description = f.read()

setup(
    name='scandb',
    version='0.4.0',
    author='Christoph Bless',
    author_email='bitbucket@cbless.de',
    url='https://bitbucket.org/cbless/scandb',
    license=' GPLv3',
    description=("Scripts to import nmap and nessus scans into an sqlite database."),
    long_description=long_description,
    long_description_content_type='text/markdown',
    packages=['scandb'],
    install_requires=[
        'argparse',
        'termcolor',
        'python-libnmap',
        'python-libnessus',
        'peewee'
    ],
    entry_points = {
        "console_scripts": [
            "scandb-vulns = scandb.vulns:vulns_cli",
            "scandb-services = scandb.services:services_cli",
            "scandb-statistics = scandb.statistics:statistics_cli",
            "scandb-importer = scandb.importer:importer"
        ]
    }
)

